/***************************************************************************
                          cbase64.cpp  -  description
                             -------------------
    begin                : Tue Jul 09 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <dclib/dcos.h>
#include <dclib/cbytearray.h>

#include "cbase64.h"

// original from mutt

static int Index_64[128] = {
-1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
-1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
-1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,62, 63,-1,-1,-1,
52,53,54,55, 56,57,58,59, 60,61,-1,-1, -1,-1,-1,-1,
-1, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
15,16,17,18, 19,20,21,22, 23,24,25,-1, -1,-1,-1,-1,
-1,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
41,42,43,44, 45,46,47,48, 49,50,51,-1, -1,-1,-1,-1
};

static char B64Chars[64] = {
'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
'8', '9', '+', ','
};

#define base64val(c) Index_64[(unsigned int)(c)]

/** */
CBase64::CBase64()
{
}

/** */
CBase64::~CBase64()
{
}

#define BAD     -1

/* raw bytes to null-terminated base 64 string */
void CBase64::Encode(unsigned char *out, const unsigned char *in, size_t len, size_t olen)
{
	while (len >= 3 && olen > 10)
	{
		*out++ = B64Chars[in[0] >> 2];
		*out++ = B64Chars[((in[0] << 4) & 0x30) | (in[1] >> 4)];
		*out++ = B64Chars[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
		*out++ = B64Chars[in[2] & 0x3f];
		olen  -= 4;
		len   -= 3;
		in    += 3;
	}

	/* clean up remainder */
	if (len > 0 && olen > 4)
	{
		unsigned char fragment;

		*out++ = B64Chars[in[0] >> 2];
		fragment = (unsigned char)((in[0] << 4) & 0x30);
		if (len > 1)
			fragment |= in[1] >> 4;
		*out++ = B64Chars[fragment];
		*out++ = (len < 2) ? '=' : B64Chars[(in[1] << 2) & 0x3c];
		*out++ = '=';
	}
	*out = '\0';
}

/* Convert '\0'-terminated base 64 string to raw bytes.
 * Returns length of returned buffer, or -1 on error */
int CBase64::Decode( CByteArray *outa, CByteArray *ina )
{
	int len = 0;
	register unsigned char digit1, digit2, digit3, digit4;
	unsigned char *in;
	unsigned char c;

	if ( (!outa) || (!ina) )
		return -1;
	if ( ina->Size() < 4 )
		return -1;
	outa->SetSize(0);

	in = ina->Data();

	do
	{
		digit1 = in[0];
		if (digit1 > 127 || base64val (digit1) == BAD)
			return -1;
		digit2 = in[1];
		if (digit2 > 127 || base64val (digit2) == BAD)
			return -1;
		digit3 = in[2];
		if (digit3 > 127 || ((digit3 != '=') && (base64val (digit3) == BAD)))
			return -1;
		digit4 = in[3];
		if (digit4 > 127 || ((digit4 != '=') && (base64val (digit4) == BAD)))
			return -1;
		in += 4;

		/* digits are already sanity-checked */
		c = (unsigned char)((base64val(digit1) << 2) | (base64val(digit2) >> 4));
		outa->Append( &c, 1);
		len++;
		if (digit3 != '=')
		{
			c = (unsigned char)(((base64val(digit2) << 4) & 0xf0) | (base64val(digit3) >> 2));
			outa->Append( &c, 1);
			len++;
			if (digit4 != '=')
			{
				c = (unsigned char)(((base64val(digit3) << 6) & 0xc0) | base64val(digit4));
				outa->Append( &c, 1);
				len++;
			}
		}
	}
	while (*in && digit4 != '=');

	return len;
}
