/***************************************************************************
                          cdownloadmanager.cpp  -  description
                             -------------------
    begin                : Don Mai 16 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cconfig.h>
#include <dclib/cdir.h>
#include <dclib/cservermanager.h>
#include <dclib/cmd5.h>

#ifndef WIN32
#include <unistd.h>
#endif

#define CHUNK_SIZE (1024*1024)

#include "cdownloadmanager.h"

extern CConfig * pDCLibConfig;
extern CServerManager * pDCLibServerManager;

CDownloadManager * pDownloadManager = 0;

CDownloadManagerInfo::CDownloadManagerInfo()
{
	eType    = DC_MESSAGE_DM_INFO;
	rate_ul  = 0;
	rate_dl  = 0;
	slot_max = 0;
	slot_use = 0;
}

/** */
CDownloadManager::CDownloadManager()
{
	lTransferID    = time(0);
	lUsedSlots     = 0;
	iShutdownState = 0;

	tDownloadQueueTimeout  = time(0);
	tUpdateTransferTimeout = time(0);

	pDownloadManager = this;

	pTransferList      = new CThreadStringList();
	pExtraUserSlotList = new CList<CExtraUserSlot>;
	pTransferWaitList  = new CThreadList<DCTransferWait>;
}

/** */
CDownloadManager::~CDownloadManager()
{
	StopListen();

	Stop();

	if (pExtraUserSlotList)
	{
		delete pExtraUserSlotList;
		pExtraUserSlotList = 0;
	}

	if(pTransferList)
	{
		delete pTransferList;
		pTransferList = 0;
	}

	if(pTransferWaitList)
	{
		delete pTransferWaitList;
		pTransferWaitList = 0;
	}
}

/** */
void CDownloadManager::SendFileInfo( DCTransferObject * TransferObject, DCTransferFileObject * TransferFileObject, bool bRemoveFile )
{
	LogThread.Lock();

	CMessageDMFileObject * fo = new CMessageDMFileObject();

	fo->sNick       = TransferObject->sNick;
	fo->sHubName    = TransferObject->sHubName;
	fo->sHubHost    = TransferObject->sHubHost;
	fo->eWaitState  = TransferObject->eState;
	fo->tTimeout    = TransferObject->tTimeout;
	fo->bRemoveFile = bRemoveFile;
	fo->iConnections = TransferObject->iConnections;

	if ( TransferFileObject )
	{
		fo->sRemoteFile      = TransferFileObject->sRemoteFile;
		fo->sLocalFile       = TransferFileObject->sLocalFile;
		fo->lSize            = TransferFileObject->lSize;
		fo->eFileState       = TransferFileObject->eState;
		fo->bMulti           = TransferFileObject->bMulti;
	}

	fo->eType = DC_MESSAGE_FILE_OBJECT;

	if ( DC_DownloadManagerCallBack(fo) == -1 )
		delete fo;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendTransferInfo( CTransfer * Transfer, bool remove )
{
	LogThread.Lock();

	CMessageDMTransferObject * to = new CMessageDMTransferObject();

	to->sTransferID      = Transfer->GetTransferID();
	to->sNick            = Transfer->GetDstNick();
	to->sHost            = Transfer->GetHost();
	to->sHubName         = Transfer->GetHubName();
	to->eState           = Transfer->GetMode();
	to->sRemoteFile      = Transfer->GetDstFilename();
	to->sRemoteFileName  = "";
	to->lSize            = Transfer->GetLength();
	to->lStartPosition   = Transfer->GetStartPosition();
	to->lCurrentPosition = Transfer->GetCurrentPosition();
	to->lEndPosition     = Transfer->GetEndPosition();

	to->lRate            = Transfer->GetTransferrate();
	to->bRemoveTransfer  = remove;

	to->eType = DC_MESSAGE_TRANSFER_OBJECT;

	if ( DC_DownloadManagerCallBack(to) == -1 )
		delete to;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendSlotInfo( CExtraUserSlot * Object )
{
	LogThread.Lock();

	CMessageDMSlotObject * so = new CMessageDMSlotObject();

	so->sNick      = Object->sNick;
	so->sHubName   = Object->sHubName;
	so->iSlots     = Object->iSlots;
	so->bPermanent = Object->bPermanent;

	so->eType = DC_MESSAGE_SLOT_OBJECT;

	if ( DC_DownloadManagerCallBack(so) == -1 )
		delete so;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendLogInfo( CString message, CTransfer * Transfer )
{
	LogThread.Lock();

	CString s = "";

	CMessageLog * log = new CMessageLog();

	if ( Transfer != 0 )
	{
		s = "[" + Transfer->GetDstNick() + "] ";
	}
	
	s += message;
	
	log->sMessage = s;
	log->eType    = DC_MESSAGE_LOG;

	if ( DC_DownloadManagerCallBack(log) == -1 )
		delete log;

	LogThread.UnLock();
}

/** */
void CDownloadManager::SendDownloadManagerInfo( CDownloadManagerInfo * dminfo )
{
	LogThread.Lock();

	bool b = FALSE;

	if ( dminfo->rate_dl != DownloadManagerInfo.rate_dl )
		b = TRUE;
	else if ( dminfo->rate_ul != DownloadManagerInfo.rate_ul )
		b = TRUE;
	else if ( dminfo->slot_max != DownloadManagerInfo.slot_max )
		b = TRUE;
	else if ( dminfo->slot_use != DownloadManagerInfo.slot_use )
		b = TRUE;

	if ( b == TRUE )
	{
		CDownloadManagerInfo * dmi = new CDownloadManagerInfo;

		dmi->rate_dl  = DownloadManagerInfo.rate_dl  = dminfo->rate_dl;
		dmi->rate_ul  = DownloadManagerInfo.rate_ul  = dminfo->rate_ul;
		dmi->slot_max = DownloadManagerInfo.slot_max = dminfo->slot_max;
		dmi->slot_use = DownloadManagerInfo.slot_use = dminfo->slot_use;

		if ( DC_DownloadManagerCallBack(dmi) == -1 )
			delete dmi;
	}

	LogThread.UnLock();
}

/** */
void CDownloadManager::FileListDone( CTransfer * Transfer )
{
	DCTransferObject * TransferObject;
	CByteArray ba;

	Lock();

	if ( Transfer->GetBuffer(&ba) != 0 )
	{
		if ( (TransferObject = pDownloadQueue.GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
		{
			TransferObject->sUserFileList.Set((char*)ba.Data(),ba.Size());

			CMessageDMFileListObject * flo = new CMessageDMFileListObject();

			flo->sNick         = TransferObject->sNick;
			flo->sHubName      = TransferObject->sHubName;
			flo->sHubHost      = TransferObject->sHubHost;
			flo->sUserFileList = TransferObject->sUserFileList;

			flo->eType = DC_MESSAGE_FILELIST_OBJECT;

			if ( DC_DownloadManagerCallBack(flo) == -1 )
				delete flo;
		}
	}

	UnLock();
}

/** */
bool CDownloadManager::CheckHash( CTransfer * Transfer )
{
	bool res = FALSE;
	CByteArray ba;
	DCTransferFileObject * TransferFileObject;
	DCFileChunkObject * FileChunkObject;
	MD5 context;

	if ( (TransferFileObject = pDownloadQueue.GetUserFileObject( Transfer->GetDstNick(), Transfer->GetHubName(), Transfer->GetDstFilename() )) != 0 )
	{
		if ( Transfer->GetBuffer(&ba) != 0 )
		{
			if ( (TransferFileObject->stHash == "") && (TransferFileObject->bMulti == TRUE) )
			{
				// calc hash
				context.update( ba.Data(), ba.Size() );
				context.finalize();

				TransferFileObject->stHash = context.hex_digest();
				printf("hash is :'%s'\n",TransferFileObject->stHash.Data());

				pDownloadQueue.pChunkList->Lock();

				if ( pDownloadQueue.pChunkList->Get( Transfer->GetSrcFilename(), (CObject *&)FileChunkObject ) == 0 )
				{
					if ( FileChunkObject->stHash == "" )
					{
						printf("Set hash ...\n");
						FileChunkObject->stHash = TransferFileObject->stHash;

						res = TRUE;
					}
					else if ( FileChunkObject->stHash == TransferFileObject->stHash )
					{
						printf("Hash ok...\n");

						res = TRUE;
					}
					else
					{
						printf("Wrong hash !!!\n");
						TransferFileObject->eState = etfsERROR;
					}
				}
				else
				{
					printf("warning file chunk object not found\n");
				}

				pDownloadQueue.pChunkList->UnLock();
			}
			else
			{
				printf("warning hash not empty or no multi download\n");
			}
		}
		else
		{
			printf("warning file object not found\n");
		}
	}
	else
	{
		printf("warning get buffer error\n");
	}

	return res;
}

/** */
int CDownloadManager::LoadQueue()
{
	int err;
	CStringList * StringList = 0;

	Lock();

	pDownloadQueue.pQueue->Clear();

	err = pDCLibConfig->LoadDCTra(pDownloadQueue.pQueue,pDownloadQueue.pChunkList);

	// send all files over the callback function
	if ( err == 0 )
	{
		while( pDownloadQueue.pQueue->Next( (CObject *&) StringList) )
		{
			DCTransferObject * TransferObject = 0;

			while( StringList->Next( (CObject *&) TransferObject) )
			{
				DCTransferFileObject * TransferFileObject = 0;

				while( TransferObject->pTransferFileList.Next( (CObject *&)TransferFileObject) )
				{
					SendFileInfo( TransferObject, TransferFileObject );
				}
			}
		}
	}

	UnLock();

	Start();

	return err;
}

/** */
int CDownloadManager::SaveQueue()
{
	int err;

	Lock();
	pDownloadQueue.pChunkList->Lock();

	err = pDCLibConfig->SaveDCTra(pDownloadQueue.pQueue,pDownloadQueue.pChunkList);

	pDownloadQueue.pChunkList->UnLock();
	UnLock();

	return err;
}

/** */
bool CDownloadManager::InitListen( CString & s )
{
	bool res = FALSE;

	if ( Listen.StartListen( pDCLibConfig->GetTCPListenPort() ) == 0 )
	{
		Listen.SetCallBackFunction( new CCallback<CDownloadManager>( this, &CDownloadManager::DM_ListenCallBack ) );

		res = TRUE;
	}
	else
	{
		s = Listen.GetSocketError();
	}

	return res;
}

/** */
void CDownloadManager::StopListen()
{
	Listen.SetCallBackFunction(0);
	Listen.StopListen();
}

/** */
void CDownloadManager::AddUserSlot( CString nick, CString hubname, int slot, bool permanent )
{
	CExtraUserSlot * ExtraUserSlot;
	Lock();

	ExtraUserSlot = 0;

	while( (ExtraUserSlot=pExtraUserSlotList->Next(ExtraUserSlot)) != 0 )
	{
		if ( (ExtraUserSlot->sNick == nick) && (ExtraUserSlot->sHubName == hubname) )
		{
			if ( (slot == 0) && (!permanent) )
			{
				ExtraUserSlot->iSlots     = slot;
				ExtraUserSlot->bPermanent = FALSE;
			}
			else
			{
				if (ExtraUserSlot->bPermanent)
					ExtraUserSlot->iSlots = 0;
				ExtraUserSlot->iSlots     += slot;
				ExtraUserSlot->bPermanent  = permanent;
			}
			break;
		}
	}

	if ( ExtraUserSlot == 0 )
	{
		ExtraUserSlot = new CExtraUserSlot();
		ExtraUserSlot->sNick      = nick;
		ExtraUserSlot->sHubName   = hubname;
		ExtraUserSlot->iSlots     = slot;
		ExtraUserSlot->bPermanent = permanent;

		pExtraUserSlotList->Add(ExtraUserSlot);
	}

	SendSlotInfo(ExtraUserSlot);

	if ( (ExtraUserSlot->iSlots == 0) && (!ExtraUserSlot->bPermanent) )
	{
		pExtraUserSlotList->Del(ExtraUserSlot);
	}

	UnLock();
}

/** */
bool CDownloadManager::CheckUserSlot( CString nick, CString )
{
	bool res = FALSE;
	CExtraUserSlot * ExtraUserSlot = 0;

	Lock();

	while( (ExtraUserSlot=pExtraUserSlotList->Next(ExtraUserSlot)) != 0 )
	{
		if ( ExtraUserSlot->sNick == nick )
		{
			if ( ExtraUserSlot->iSlots > 0 )
			{
				ExtraUserSlot->iSlots--;

				SendSlotInfo(ExtraUserSlot);

				if ( ExtraUserSlot->iSlots == 0 )
				{
					pExtraUserSlotList->Del(ExtraUserSlot);
				}

				res = TRUE;
			}
			else if ( ExtraUserSlot->bPermanent )
			{
				res = TRUE;
			}

			break;
		}
	}

	UnLock();

	return res;
}

/** disconnect all transfers for a clean shutdown */
void CDownloadManager::Shutdown()
{
	CObject *obj;
	CTransfer * Transfer;

	StopListen();

	iShutdownState = 1;

	SendLogInfo("Shutdown download manager ...\n");

	pTransferList->Lock();

	printf("Running Transfers: %ld\n",pTransferList->Count());

	obj = 0;
	while( pTransferList->Next( (CObject*&)obj ) )
	{
		Transfer = (CTransfer*)obj;
		Transfer->Disconnect(TRUE);
	}

	pTransferList->UnLock();
}

/** */
bool CDownloadManager::TransferClose( CString transferid )
{
	CObject *obj;
	CTransfer * Transfer;
	bool found=FALSE;

	pTransferList->Lock();

	if ( pTransferList->Get( transferid, (CObject*&)obj) == 0 )		
	{
		Transfer = (CTransfer*)obj;
		Transfer->Disconnect(TRUE);
		found = TRUE;
	}

	pTransferList->UnLock();

	return found;
}

/** */
bool CDownloadManager::TransferCancel( CString nick, CString host, CString file )
{
	CObject * obj;
	CTransfer * Transfer;
	bool found=FALSE;

	pTransferList->Lock();

	obj = 0;

	while( pTransferList->Next( (CObject*&)obj ) )
	{
		Transfer = (CTransfer*)obj;
		if ( (Transfer->GetDstNick() == nick) &&	
		     (Transfer->GetHost() == host) &&
		     (Transfer->GetDstFilename() == file) )
		{
			Transfer->SendString("$Cancel");
			found = TRUE;
			break;
		}
	}

	pTransferList->UnLock();

	return found;
}

/** */
eDirection CDownloadManager::TransferDirection( CString transferid )
{
	CObject * obj;
	CTransfer * Transfer;
	eDirection direction=edNONE;

	pTransferList->Lock();

	obj = 0;

	if ( pTransferList->Get( transferid, (CObject*&)obj ) == 0 )
	{
		Transfer = (CTransfer*)obj;
		direction = Transfer->GetSrcDirection();
	}

	pTransferList->UnLock();

	return direction;
}

/** */
bool CDownloadManager::TransferSetRate( CString transferid, ulonglong rate )
{
	CObject * obj;
	CTransfer * Transfer;
	bool found=FALSE;

	pTransferList->Lock();

	if ( pTransferList->Get( transferid, (CObject*&)obj ) == 0 )
	{
		Transfer = (CTransfer*)obj;
		Transfer->SetRate(rate);
		found=TRUE;
	}

	pTransferList->UnLock();

	return found;
}

/** */
ulonglong CDownloadManager::TransferGetRate( CString transferid )
{
	CObject * obj;
	CTransfer * Transfer;
	ulonglong rate = 0;

	pTransferList->Lock();

	obj = 0;

	if ( pTransferList->Get( transferid, (CObject*&)obj) == 0 )
	{
		Transfer = (CTransfer*)obj;
		rate = Transfer->GetRate();
	}

	pTransferList->UnLock();

	return rate;
}

/** */
bool CDownloadManager::GetUserFileInfo( CString nick, CString hubname, CString /*hubhost*/, CString remotefile, CUserFileInfo * UserFileInfo )
{
	bool res = FALSE;

	if (!UserFileInfo)
	{
		return res;
	}

	Lock();

	DCTransferObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) != 0 )
	{
		UserFileInfo->eWaitState    = TransferObject->eState;
		UserFileInfo->sUserFileList = TransferObject->sUserFileList;

		if ( remotefile != "" )
		{
			if ( TransferObject->pTransferFileList.Get( remotefile, (CObject *&) TransferFileObject ) == 0 )
			{
				UserFileInfo->eFileState = TransferFileObject->eState;
				UserFileInfo->sLocalFile = TransferFileObject->sLocalFile;

				res = TRUE;
			}
		}
		else
		{
			res = TRUE;
		}
	}

	UnLock();

	return res;
}

/** */
DCFileChunkObject * CDownloadManager::GetFileChunk( CString file )
{
	Lock();

	DCFileChunkObject * FileChunkObject1, * FileChunkObject2=0;

	if ( (FileChunkObject1 = pDownloadQueue.GetFileChunkObject(file)) != 0 )
	{
		FileChunkObject2 = new DCFileChunkObject(FileChunkObject1);
	}

	UnLock();

	return FileChunkObject2;
}

/** */
void CDownloadManager::TryConnect( CString nick, CString hubname )
{
	Lock();

	DCTransferObject * TransferObject;

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) != 0 )
	{
		TransferObject->tTimeout = 0;
	}

	UnLock();
}

/** */
bool CDownloadManager::UpdateServer( CString nick, CString hubname )
{
	Lock();

	bool res = FALSE;
	DCConfigHubItem hubitem;
	DCTransferObject * TransferObject;

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( (res = pDCLibConfig->GetPublicHub( hubname, &hubitem )) == TRUE )
			TransferObject->sHubHost = hubitem.sHost;
		else if ( (res = pDCLibConfig->GetBookmarkHub( hubname, &hubitem )) == TRUE )
			TransferObject->sHubHost = hubitem.sHost;

		if ( res == TRUE )
			SendFileInfo( TransferObject );
	}

	UnLock();

	return res;
}

/** */
bool CDownloadManager::AddTransfer( CString nick, CString userhost, CString hubname, CString hubhost )
{
	bool res = FALSE;
	DCTransferWait * TransferWait = 0;

	printf("Add wait transfer for: '%s' '%s' '%s'\n",nick.Data(),hubname.Data(),hubhost.Data());

	pTransferWaitList->Lock();

	while ( (TransferWait = pTransferWaitList->Next(TransferWait)) != 0 )
	{
		if ( TransferWait->sHubName == hubname )
		{
			if ( ((TransferWait->sNick == nick) && (TransferWait->sNick != "")) ||
			     ((TransferWait->sUserHost == userhost) && (TransferWait->sUserHost != "")))
			{
				// reset timeout
				TransferWait->tTimeout = time(0);

				break;
			}
		}
	}

	if ( TransferWait == 0 )
	{
		TransferWait = new DCTransferWait();
		TransferWait->sNick     = nick;
		TransferWait->sUserHost = userhost;
		TransferWait->sHubName  = hubname;
		TransferWait->sHubHost  = hubhost;
		TransferWait->tTimeout  = time(0);

		pTransferWaitList->Add(TransferWait);
	}

	res = TRUE;

	pTransferWaitList->UnLock();

	return res;
}

/** */
void CDownloadManager::AddTransfer( CString host, int port, CString hubname, CString hubhost )
{
	pTransferList->Lock();

	CTransfer * Transfer = new CTransfer();

	Transfer->SetTransferID( GetNewTransferID() );
	Transfer->SetNick( pDCLibConfig->GetNick() );

	Transfer->SetHubName(hubname);
	Transfer->SetHubHost(hubhost);

	Transfer->SetHost( host, port );
	Transfer->SetRate( pDCLibConfig->GetMaxTransferrate() );

	AddTransfer( "", Transfer->GetHost(), hubname, hubhost );

	pTransferList->Add( Transfer->GetTransferID(), Transfer );
	Transfer->SetCallBackFunction( new CCallback<CDownloadManager>( this, &CDownloadManager::DM_TransferCallBack ) );
	Transfer->Connect();

	pTransferList->UnLock();
}

/** */
CString CDownloadManager::GetNewTransferID()
{
	lTransferID++;

	if (lTransferID==0)
		lTransferID++;

	return CString().setNum(lTransferID);
}

/**	0: file not in the queue
	1: file allready in the queue by same nick und hub
	2: local file with same size allready in the queue (multi download)
	3: local file with same size allready in the queue without multi (no multi download)
	4: local file with not the same size allready in the queue (multi download error)
*/
int CDownloadManager::CheckFile( CString nick, CString hubname, CString hubhost,
				CString remotename, CString localname,
				CString localpath, CString localrootpath,
				eltMedium, ulonglong size )
{
	Lock();

	CDir dir;
	CString sfile;
	int res = 0;
	DCTransferObject * TransferObject;
	DCTransferFileObject * TransferFileObject = 0;

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( (TransferFileObject = pDownloadQueue.GetUserFileObject( nick, hubname, remotename )) != 0 )
		{
			res = 1;
		}
	}

	// check if a local file allready exist
	if ( (res == 0) && (remotename != CString(DC_USER_FILELIST)) )
	{
		if ( localrootpath == "" )
		{
			sfile = pDCLibConfig->GetDownloadFolder();
		}
		else
		{
			sfile = localrootpath;
		}

		sfile = sfile + "/" + localpath + "/" + localname;
		sfile = dir.SimplePath(sfile);

		DCFileChunkObject * FileChunkObject;

		pDownloadQueue.pChunkList->Lock();

		if ( pDownloadQueue.pChunkList->Get( sfile, (CObject *&)FileChunkObject ) == 0 )
		{
			if ( FileChunkObject->lSize == size )
			{
				if ( FileChunkObject->bMulti == TRUE )
				{
					res = 2;
				}
				else
				{
					res = 3;
				}
			}
			else
			{
				res = 4;
			}
		}

		pDownloadQueue.pChunkList->UnLock();
	}

	UnLock();

	return res;
}

/** */
void CDownloadManager::AddWaitTransfer( CString nick, CString hubname, CString hubhost,
				CString remotename, CString localname,
				CString localpath, CString localrootpath,
				eltMedium medium, ulonglong size,
				ulonglong startposition,
				bool multi )
{
	CDir dir;
	CString sfile;
	DCTransferObject * TransferObject;
	DCTransferFileObject * TransferFileObject = 0;

	printf("add wait transfer: %s %s %s %s %d %s %s %s %llu %llu\n",
		nick.Data(),
		remotename.Data(),
		hubname.Data(),
		hubhost.Data(),
		medium,
		localname.Data(),
		localpath.Data(),
		localrootpath.Data(),
		startposition,
		size );

	Lock();

	CStringList * StringList = pDownloadQueue.GetUserHubList( nick );

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) == 0 )
	{
		TransferObject = new DCTransferObject();

		TransferObject->sNick    = nick;
		TransferObject->sHubHost = hubhost;
		TransferObject->sHubName = hubname;
		TransferObject->eState   = etwsIDLE;
		TransferObject->tTimeout = 0;
		TransferObject->iConnections = 0;

		if ( StringList == 0 )
		{
			StringList = new CStringList();
			pDownloadQueue.pQueue->Add( nick, StringList );
		}

		StringList->Add( hubname, TransferObject );

		printf("transfer add\n");
	}
	else
	{
		TransferFileObject = pDownloadQueue.GetUserFileObject( nick, hubname, remotename );
	}

	// file not found for the nick/hub
	if ( TransferFileObject == 0 )
	{
		TransferFileObject = new DCTransferFileObject();

		TransferFileObject->eState           = etfsNONE;
		TransferFileObject->lSize            = size;
		TransferFileObject->bMulti           = multi;
		TransferFileObject->eMedium          = medium;
		TransferFileObject->sRemoteFile      = remotename;

		if ( localrootpath == "" )
		{
			sfile = pDCLibConfig->GetDownloadFolder();
		}
		else
		{
			sfile = localrootpath;
		}

		sfile = sfile + "/" + localpath + "/" + localname;
		sfile = dir.SimplePath(sfile);

		TransferFileObject->sLocalFile     = sfile;
		TransferFileObject->sLocalPath     = localpath;
		TransferFileObject->sLocalFileName = localname;

		TransferObject->pTransferFileList.Add( remotename, TransferFileObject );

		printf("file add '%s' %ld\n",sfile.Data(),TransferObject->pTransferFileList.Count());

		pDownloadQueue.pChunkList->Lock();

		DCFileChunkObject * FileChunkObject;
		if ( remotename == CString(DC_USER_FILELIST) )
		{
			printf("no chunk for userlists\n");
		}
		else if ( pDownloadQueue.pChunkList->Get( sfile, (CObject *&)FileChunkObject ) != 0 )
		{
			FileChunkObject = new DCFileChunkObject();

			FileChunkObject->sLocalFile = sfile;
			FileChunkObject->stHash     = "";
			FileChunkObject->bMulti     = multi;
			FileChunkObject->lSize      = size;
			FileChunkObject->lSizeDone  = 0;

			// create the first unfinished chunk
			DCChunkObject * ChunkObject = new DCChunkObject();
			ChunkObject->lStart = startposition;
			ChunkObject->lEnd   = size;
			FileChunkObject->pChunks.Add(ChunkObject);

			pDownloadQueue.pChunkList->Add( sfile, FileChunkObject );

			printf("add file chunk object\n");
		}
		else
		{
			printf("file chunk object found\n");
		}

		pDownloadQueue.pChunkList->UnLock();
	}
	else
	{
		printf("file found ...\n");
	}

	SendFileInfo( TransferObject, TransferFileObject );

	UnLock();
}

/** */
bool CDownloadManager::RemoveQueue( CString nick, CString hubname, CString remotefile )
{
	bool res = FALSE;
	DCTransferFileObject * TransferFileObject = 0;
	DCTransferObject * TransferObject = 0;

	Lock();

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( nick, hubname )) != 0 )
	{
		if ( remotefile != "" )
		{
			if ( (TransferFileObject = pDownloadQueue.GetUserFileObject( nick, hubname, remotefile )) != 0 )
			{
				if ( TransferFileObject->eState != etfsTRANSFER )
				{
					res = TRUE;
				}
			}
		}
		else
		{
			if ( TransferObject->eState != etwsRUN )
			{
				res = TRUE;
			}
		}

		if ( res == TRUE )
		{
			SendFileInfo( TransferObject, TransferFileObject, TRUE );
		}
	}

	if ( res == TRUE )
	{
		res = pDownloadQueue.DelUserFileObject( nick, hubname, remotefile );
	}

	UnLock();

	return res;
}

/** */
bool CDownloadManager::CheckWaitTransfer( CTransfer * Transfer )
{
	Lock();

	bool res = FALSE;
	DCTransferObject * TransferObject;
	DCTransferWait * TransferWait = 0;

	printf("CheckWaitTransfer I : %s on %s\n",
		Transfer->GetDstNick().Data(),
		Transfer->GetHubName().Data() );

	// search in the wait list ...
	pTransferWaitList->Lock();

	while ( (TransferWait = pTransferWaitList->Next(TransferWait)) != 0 )
	{
		// search for the nick ...
		if ( ((TransferWait->sNick == Transfer->GetDstNick()) && (TransferWait->sNick != "")) ||
		     ((TransferWait->sUserHost == Transfer->GetHost()) && (TransferWait->sUserHost != "")) )
		{
			Transfer->SetHubName(TransferWait->sHubName);
			Transfer->SetHubHost(TransferWait->sHubHost);

			break;
		}
	}

	pTransferWaitList->UnLock();

	printf("CheckWaitTransfer II: %s on %s [%d]\n",
		Transfer->GetDstNick().Data(),
		Transfer->GetHubName().Data(),(TransferWait!=0) );

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		printf("Waiting: %s on %s %s\n",
			TransferObject->sNick.Data(),
			TransferObject->sHubName.Data(),
			TransferObject->sHubHost.Data() );

		if ( TransferObject->eState == etwsWAIT )
		{
			printf("wait found ...\n");

			TransferObject->eState = etwsRUN;
			TransferObject->iConnections++;

			// waiting transfer found ...
			res = TRUE;

			SendFileInfo( TransferObject );
		}
	}

	if ( (res == FALSE) && (TransferWait == 0) )
	{
		printf("Warning: no wait transfer found for '%s'\n",Transfer->GetDstNick().Data());
		Transfer->Disconnect(TRUE);
	}

	if ( TransferWait != 0 )
	{
		pTransferWaitList->Lock();
		pTransferWaitList->Del(TransferWait);
		pTransferWaitList->UnLock();
	}

	UnLock();

	return res;
}

/** */
bool CDownloadManager::ChangeDirection( CTransfer * Transfer )
{
	bool res = FALSE;
	DCTransferObject * TransferObject;

	Lock();

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		printf("Waiting: %s on %s %s\n",
			TransferObject->sNick.Data(),
			TransferObject->sHubName.Data(),
			TransferObject->sHubHost.Data() );

		// change to upload ...
		if ( Transfer->GetSrcDirection() == edDOWNLOAD )
		{
			if ( TransferObject->eState == etwsRUN )
			{
				TransferObject->iConnections--;
				if ( TransferObject->iConnections == 0 )
					TransferObject->eState = etwsIDLE;
				SendFileInfo( TransferObject );

				printf("change transfer -> upload ...\n");

				res = TRUE;
			}
			else
			{
				printf("can't change transfer upload ...\n");
			}
		}
	}

	UnLock();

	return res;
}

/** */
bool CDownloadManager::SetNextFile( CTransfer * Transfer )
{
	bool res;

	Lock();
	res = SetFile(Transfer);
	UnLock();

	return res;
}

/** */
bool CDownloadManager::SetDirection( CTransfer * Transfer )
{
	int i;
	bool res = FALSE;

	// check incoming transfer limit
	i = pDCLibConfig->GetMaxUpload();

	if ( Transfer->GetSrcDirection() == edUPLOAD )
	{
		if ( i == 0 )
		{
			res = TRUE;
		}
		else if ( i > lUsedSlots )
		{
			res = TRUE;
		}
		else
		{
			res = CheckUserSlot( Transfer->GetDstNick(), Transfer->GetHubName() );
		}

		if (res)
		{
			lUsedSlots++;
		}
	}
	else
	{
		res = TRUE;
	}

	return res;
}

/** */
bool CDownloadManager::GetNextChunk( CString file, ulonglong * lstart, ulonglong * lend )
{
	bool res = FALSE;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co, *co1;
	ulonglong size;
	bool split;

	printf("get the next chunk for '%s'\n",file.Data());

	if ( (FileChunkObject = pDownloadQueue.GetFileChunkObject(file)) != 0 )
	{
		// first sort one element from the list ...
		if ( (co = FileChunkObject->pChunks.Next(0)) != 0 )
		{
			FileChunkObject->pChunks.Remove(co);

			if ( co->bLocked == FALSE )
			{
				ChunkObject = 0;
				while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
				{
					if ( ChunkObject->bLocked == FALSE )
					{
						if ( co->lEnd == ChunkObject->lStart )
						{
							co->lEnd = ChunkObject->lEnd;
							FileChunkObject->pChunks.Del(ChunkObject);
							printf("chunk list optimized\n");
							break;
						}
					}
				}
			}

			FileChunkObject->pChunks.Add(co);
		}

		co = 0;
		size = 0;
		ChunkObject = 0;
		split = TRUE;

		// get the biggest free chunk
		while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
		{
			if ( ChunkObject->bLocked == FALSE )
			{
				if ( (ChunkObject->lEnd - ChunkObject->lStart) > size )
				{
					co   = ChunkObject;
					size = (ChunkObject->lEnd - ChunkObject->lStart);
				}
			}
		}

		// found a free chunk
		if ( co != 0 )
		{
			split = FALSE;
			ChunkObject = 0;

			// check if a running chunk for this chunk
			while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
			{
				if ( ChunkObject->bLocked == TRUE )
				{
					if ( ChunkObject->lEnd == co->lStart )
					{
						split = TRUE;
					}
				}
			}

			ChunkObject = co;

			if ( (ChunkObject->lEnd - ChunkObject->lStart) > CHUNK_SIZE )
			{
				// create a new chunk
				co = new DCChunkObject();
	
				// split chunk
				if ( (FileChunkObject->bMulti == FALSE) || (ChunkObject->lStart == 0) || (split == FALSE) )
				{
					co->lStart  = ChunkObject->lStart;
					co->lEnd    = ChunkObject->lStart+CHUNK_SIZE;
					ChunkObject->lStart += CHUNK_SIZE;
				}
				else
				{
					printf("CHUNK SET 1: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);

					co->lStart  = ChunkObject->lStart+((ChunkObject->lEnd - ChunkObject->lStart)/2);

					if ( (ChunkObject->lEnd-co->lStart) > CHUNK_SIZE )
					{
						co->lEnd = co->lStart+CHUNK_SIZE;
						co1 = new DCChunkObject();
						co1->lStart = co->lEnd;
						co1->lEnd   = ChunkObject->lEnd;
						FileChunkObject->pChunks.Add(co1);
						printf("CHUNK SET 2: %llu %llu\n",co1->lStart,co1->lEnd);
					}
					else
					{
						co->lEnd = ChunkObject->lEnd;
					}

					ChunkObject->lEnd = co->lStart;

					printf("CHUNK SET 3: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);
					printf("CHUNK SET 4: %llu %llu\n",co->lStart,co->lEnd);
				}

				co->bLocked = TRUE;

				FileChunkObject->pChunks.Add(co);

				*lstart = co->lStart;
				*lend   = co->lEnd;

				printf("NEW CHUNK SPLIT/LOCKED: %llu %llu\n",*lstart,*lend);

				res = TRUE;
			}
			else
			{
				ChunkObject->bLocked = TRUE;

				*lstart = ChunkObject->lStart;
				*lend   = ChunkObject->lEnd;

				printf("NEW CHUNK LOCKED: %llu %llu\n",*lstart,*lend);

				res = TRUE;
			}
		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	return res;
}

/** 0: not found; 1: found and update; 2: no more chunks (file download ok) */
int CDownloadManager::UpdateChunk( CString file, ulonglong lstart, ulonglong lend, ulonglong lcurrent )
{
	int res = 0;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co;

	printf("update chunk for '%s'\n",file.Data());

	pDownloadQueue.pChunkList->Lock();

	if ( (FileChunkObject = pDownloadQueue.GetFileChunkObject(file)) != 0 )
	{
		ChunkObject = 0;

		// search the chunk
		while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
		{
			if ( (ChunkObject->lStart == lstart) && (ChunkObject->lEnd == lend) )
			{
				// chunk found
				res = 1;

				if ( ChunkObject->bLocked == FALSE )
				{
					printf("warning wrong chunk state\n");
				}

				ChunkObject->bLocked = FALSE;

				if ( lstart != lcurrent )
				{
					// update size done
					FileChunkObject->lSizeDone += (lcurrent-lstart);

					printf("FILESTATE: %llu %llu\n",FileChunkObject->lSizeDone,FileChunkObject->lSize);

					if ( lcurrent == lend )
					{
						FileChunkObject->pChunks.Del(ChunkObject);
						ChunkObject = 0;
					}
					else
					{
						ChunkObject->lStart = lcurrent;
					}

					if ( FileChunkObject->lSizeDone == FileChunkObject->lSize )
					{
						// remove the complete chunk from the list
						pDownloadQueue.pChunkList->Del(file);
						res = 2;
						break;
					}
				}

				if ( ChunkObject != 0 )
				{
					co = 0;
					while ( (co=FileChunkObject->pChunks.Next(co)) != 0 )
					{
						if ( ChunkObject->lEnd == co->lStart )
						{
							if ( co->bLocked == FALSE )
							{
								co->lStart = ChunkObject->lStart;
								FileChunkObject->pChunks.Del(ChunkObject);
								ChunkObject = co;
								printf("CHUNK FIX1: %llu %llu\n",ChunkObject->lStart,ChunkObject->lEnd);
							}
							break;
						}
					}

					co = 0;
					while ( (co=FileChunkObject->pChunks.Next(co)) != 0 )
					{
						if ( ChunkObject->lStart == co->lEnd )
						{
							if ( co->bLocked == FALSE )
							{
								co->lEnd = ChunkObject->lEnd;
								FileChunkObject->pChunks.Del(ChunkObject);
								printf("CHUNK FIX2: %llu %llu\n",co->lStart,co->lEnd);
							}
							break;
						}
					}
				}

				break;
			}
		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	pDownloadQueue.pChunkList->UnLock();

	return res;
}

/** */
bool CDownloadManager::GetNewChunkEnd( CString file, ulonglong lstart, ulonglong lend, ulonglong lcurrent, ulonglong * lnstart, ulonglong *lnend )
{
	pDownloadQueue.pChunkList->Lock();

	bool res = FALSE;
	DCFileChunkObject * FileChunkObject;
	DCChunkObject * ChunkObject, *co1=0,*co2=0;

	if ( (FileChunkObject = pDownloadQueue.GetFileChunkObject(file)) != 0 )
	{
		ChunkObject = 0;

		// check if no multi download ...
//		if ( FileChunkObject->bMulti == TRUE )
//		{
			while ( (ChunkObject=FileChunkObject->pChunks.Next(ChunkObject)) != 0 )
			{
				if ( (ChunkObject->lStart == lstart) && (ChunkObject->lEnd == lend) )
				{
					co1 = ChunkObject;

					if ( co2 )
						break;
				}
				else if ( ChunkObject->lStart == lend )
				{
					// the chunk after the original chunk is locked
					if ( ChunkObject->bLocked == TRUE )
					{
						break;
					}
					else
					{
						co2 = ChunkObject;
        	
						if ( co1 )
							break;
					}
				}
			}

			// found start and next chunk
			if ( (co1 != 0) && (co2 != 0) )
			{
				printf("get new chunk end for '%s'\n",file.Data());

				// if chunk > max chunk split chunk
				if ( (co2->lEnd-co2->lStart) > CHUNK_SIZE )
				{
					co2->lStart += CHUNK_SIZE;
					co1->lEnd   += CHUNK_SIZE;
				}
				else
				{
					co1->lEnd = co2->lEnd;
					FileChunkObject->pChunks.Del(co2);
				}

				if ( (lcurrent-lstart) > 0 )
				{
					FileChunkObject->lSizeDone += (lcurrent-lstart);
					co1->lStart = lcurrent;
				}

				*lnstart = co1->lStart;
				*lnend   = co1->lEnd;

				printf("new chunk end set %llu -> %llu\n",lend,*lnend);

				res = TRUE;
			}
//		}
	}
	else
	{
		printf("warning file not found in the chunk list\n");
	}

	pDownloadQueue.pChunkList->UnLock();

	return res;
}

/** */
bool CDownloadManager::SetFile( CTransfer * Transfer )
{
	bool res;
	DCTransferObject * TransferObject;
	ulonglong lstart,lend;

	res = FALSE;

	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		if ( TransferObject->pTransferFileList.Count() > 0 )
		{
			DCTransferFileObject * TransferFileObject = 0;

			while( TransferObject->pTransferFileList.Next((CObject*&)TransferFileObject) )
			{
				if ( TransferFileObject->eState == etfsNONE )
				{
					printf("set file: '%s'\n",TransferFileObject->sRemoteFile.Data());

					CString sPath="",sLocalPath="",sFile="";
					CDir dir;

					if ( TransferFileObject->eMedium == eltFILE )
					{
						sFile = TransferFileObject->sLocalFile;

						printf("DEBUG: file: '%s'\n",sFile.Data());

						// extract file
						int pos = sFile.FindRev(DIRSEPARATOR);
						if( pos != -1 )
						{
							sPath = sFile.Left(pos);
						}

						printf("DEBUG: path: '%s'\n", sPath.Data());

						// create the path
						if ( dir.CreatePath(sPath) == FALSE )
						{
							TransferFileObject->eState = etfsERROR;
							SendFileInfo( TransferObject, TransferFileObject );
							SendLogInfo( "Create path failed: " + sPath, Transfer );
							printf("DEBUG: create path failed: '%s'\n", sPath.Data());
						}
						else
						{
							printf("DOWNLOAD: '%s' %llu '%s'\n",
								TransferFileObject->sRemoteFile.Data(),
								TransferFileObject->lSize,
								sFile.Data());

								res = TRUE;
						}
					}
					else
					{
						res = TRUE;
					}

					if ( res == TRUE )
					{
						// first we create the hash over the first 1MB of a multi download file in a buffer
						if ( (TransferFileObject->bMulti == TRUE) &&
						     (TransferFileObject->stHash == ""))
						{
							printf("create the hash for the file\n");

							Transfer->SetMedium(eltBUFFER);

							lstart = 0;
							lend   = 10*1024;
						}
						else // get the next chunk of the file
						{
							if ( TransferFileObject->sRemoteFile == CString(DC_USER_FILELIST) )
							{
								lstart = 0;
								lend   = 0;
							}
							else if ( GetNextChunk( TransferFileObject->sLocalFile, &lstart, &lend ) == FALSE )
							{
								// no more chunks
								printf("no more chunks ...\n");
								continue;
							}

							Transfer->SetMedium(TransferFileObject->eMedium);
						}

						Transfer->SetDone(0);

						// mark file as transfered
						TransferFileObject->eState = etfsTRANSFER;

						Transfer->StartDownload(TransferFileObject->sRemoteFile,lstart,lstart,lend,TransferFileObject->lSize,sFile);
					}
					else
					{
						continue;
					}

					SendFileInfo( TransferObject, TransferFileObject );

					break;
				}
				else
				{
//					printf("state file: '%s' %d\n",TransferFileItem->sRemoteName.Data(),TransferFileItem->eState);
				}
			}
		}
	}

	if ( res == FALSE )
	{
		// set disconnect timeout ... save remote slot ;-)
		if ( Transfer->GetDone() != 2 )
		{
			Transfer->SetStartTime(time(0));
			Transfer->SetDone(2);
		}
	}

	return res;
}

/** */
void CDownloadManager::UpdateFileState( CTransfer * Transfer, eTransferFileState eState )
{
	DCTransferObject * TransferObject;
	DCTransferFileObject * TransferFileObject;

	Lock();

	printf("updatefile\n");
	if ( (TransferObject = pDownloadQueue.GetUserTransferObject( Transfer->GetDstNick(), Transfer->GetHubName() )) != 0 )
	{
		TransferFileObject = pDownloadQueue.GetUserFileObject( Transfer->GetDstNick(), Transfer->GetHubName(), Transfer->GetDstFilename() );

		if ( TransferFileObject != 0 )
		{
			if ( TransferFileObject->eState != etfsTRANSFER )
			{
				printf("warning, wrong state in updatefile\n");
			}
			// if this a check hash dl ?
			else if ( (Transfer->GetMedium() == eltBUFFER) && (TransferFileObject->bMulti == TRUE) )
			{
				printf("updatefile hash\n");
				if ( eState == etfsNONE )
				{
					if ( Transfer->GetCurrentPosition() == Transfer->GetEndPosition() )
					{
						if ( CheckHash(Transfer) == TRUE )
						{
							SendLogInfo( "Hash ok '" + TransferFileObject->sRemoteFile + "'", Transfer );
							TransferFileObject->eState = etfsNONE;
						}
						else
						{
							SendLogInfo( "Hash failed '" + TransferFileObject->sRemoteFile + "'", Transfer );
							TransferFileObject->eState = etfsERROR;
						}
					}
					else
					{
						TransferFileObject->eState = eState;
					}
				}
				else
				{
					TransferFileObject->eState = eState;
				}

				SendFileInfo( TransferObject, TransferFileObject );
			}
			else
			{
				printf("updatefile normal\n");

				int state=0;

				TransferFileObject->eState = eState;

				// check if mylist.dclist done
				if ( TransferFileObject->sRemoteFile == CString(DC_USER_FILELIST) )
				{
					if ( (Transfer->GetLength() != 0) &&
					     (Transfer->GetLength() == Transfer->GetCurrentPosition()) )
					{
						state = 2;
					}
				}
				else
				{
					state = UpdateChunk( TransferFileObject->sLocalFile, Transfer->GetStartPosition(),
						Transfer->GetEndPosition(), Transfer->GetCurrentPosition() );
				}

				TransferFileObject->lSize = Transfer->GetLength();

				// download finished
				if ( state == 2 )
				{
					SendLogInfo( "Transfer done '" + TransferFileObject->sRemoteFile + "'", Transfer );
					SendFileInfo( TransferObject, TransferFileObject, TRUE );

					// move the file to the finished path
					if ( (TransferFileObject->eMedium == eltFILE) && // only files
					     (pDCLibConfig->GetDownloadFinishedFolder() != "") && // only if dffolder set
					     (TransferFileObject->sRemoteFile != CString(DC_USER_FILELIST)) && // only if this not a filelist
					     (TransferFileObject->sLocalFile.Find(pDCLibConfig->GetDownloadFolder().Data()) == 0) ) // only if the download in the download folder
					{
						CString s;

						CDir dir(pDCLibConfig->GetDownloadFinishedFolder());

						if ( dir.CreatePath(TransferFileObject->sLocalPath) == TRUE )
						{
#ifdef WIN32
							s  = "move ";
#else
							s  = "mv -f ";
#endif
							s += "'" + TransferFileObject->sLocalFile + "' ";
							s += "'" + pDCLibConfig->GetDownloadFinishedFolder() + DIRSEPARATOR;
							s += TransferFileObject->sLocalPath + DIRSEPARATOR + "'";
//							s += TransferFileObject->sLocalFileName;

							printf("move file '%s'\n",s.Data());
						
							if ( system(s.Data()) != 0 )
							{
								printf("move failed !\n");
							}
						}
						else
						{
							printf("move failed (create path)!\n");
						}
					}

					// remove the file from the wait queue
					TransferObject->pTransferFileList.Del(TransferFileObject->sRemoteFile);

					// TODO: remove all other multi download files ...
				}
				else
				{
					SendFileInfo( TransferObject, TransferFileObject );
				}
			}
		}
	}

	UnLock();
}

/** */
void CDownloadManager::UpdateTransferList( time_t ttimeout )
{
	long count;
	CTransfer * pTransfer;
	CObject *obj,*oobj;
	CDownloadManagerInfo dmi;

	pTransferList->Lock();

	count = pTransferList->Count();

	// check if shutdown ready
	if ( (iShutdownState == 1) && (count == 0) )
	{
		iShutdownState = 2;
	}

	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		// update slots
		dmi.slot_use = lUsedSlots;
		dmi.slot_max = pDCLibConfig->GetMaxUpload();
	}

	if ( count > 0 )
	{
		obj = oobj = 0;

		while( pTransferList->Next( (CObject*&)obj) )
		{
			pTransfer = (CTransfer*)obj;

			// refresh transfer
			pTransfer->Thread(0);

			if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
			{
				// done==2 set new file or disconnect
				if ( (pTransfer->GetDone() == 2) && (iShutdownState == 0) )
				{
					if ( SetFile(pTransfer) == FALSE )
					{
						// no new file ...
						if ( pTransfer->GetStartTime() != 0 )
						{
							if ( (ttimeout-pTransfer->GetStartTime()) >= 60 )
							{
								pTransfer->SetStartTime(0);
								pTransfer->Disconnect(TRUE);
							}
						}
					}

					oobj = obj;
				}
				else if ( pTransfer->GetDone() == 1 )
				{
					if ( pTransfer->GetDstDirection() == edDOWNLOAD )
					{
						lUsedSlots--;
					}

					SendTransferInfo( pTransfer, TRUE );

					// update internal transfer state for this user/hubname
					// set to idle (this work only with 1 transfer per user/hubname ...)
					DCTransferObject * TransferObject;

					if ( pTransfer->GetSrcDirection() == edDOWNLOAD )
					{
						if ( (TransferObject = pDownloadQueue.GetUserTransferObject( pTransfer->GetDstNick(), pTransfer->GetHubName() )) != 0 )
						{
							TransferObject->iConnections--;
							if ( TransferObject->iConnections == 0 )
								TransferObject->eState = etwsIDLE;
							TransferObject->tTimeout = ttimeout;
							SendFileInfo( TransferObject );
						}
					}

					pTransferList->Del( pTransfer->GetTransferID() );

					obj = oobj;
				}
				else if ( iShutdownState == 0 )
				{
					if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
					{
						SendTransferInfo(pTransfer);
					}

					if ( pTransfer->GetSrcDirection() == edUPLOAD )
					{
						dmi.rate_ul += pTransfer->GetTransferrate();
					}
					else if ( pTransfer->GetSrcDirection() == edDOWNLOAD )
					{
						dmi.rate_dl += pTransfer->GetTransferrate();
					}
					oobj = obj;
				}
				else
				{
					oobj = obj;
				}
			}
		}
	}

	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		SendDownloadManagerInfo(&dmi);
	}

	pTransferList->UnLock();
}

/** */
void CDownloadManager::UpdateQueueList( time_t ttimeout )
{
	int i;
	CStringList * StringList = 0;

	Lock();

	while( pDownloadQueue.pQueue->Next( (CObject *&) StringList) )
	{
		DCTransferObject * TransferObject = 0;

		while( StringList->Next( (CObject *&) TransferObject) )
		{
			if ( (TransferObject->pTransferFileList.Count() == 0) && (TransferObject->eState == etwsIDLE) )
			{
				// empty filelist
				SendFileInfo( TransferObject, 0, TRUE );
				StringList->Del(TransferObject->sHubName);
				break;
			}
			else if ( (TransferObject->eState == etwsWAIT) && (iShutdownState == 0) )
			{
				// resend connection request on wait timeout
				if ( (ttimeout-TransferObject->tTimeout) >= pDCLibConfig->GetTransferResendTimeout() )
				{
					// set timeout
					TransferObject->eState = etwsIDLE;
					TransferObject->tTimeout = ttimeout;
					SendFileInfo( TransferObject );
				}
			}
			else if ( (iShutdownState == 0) &&
				( (TransferObject->eState == etwsIDLE) ||
				  (TransferObject->eState == etwsHUBOFFLINE) ||
				  (TransferObject->eState == etwsUSEROFFLINE) ||
				  (TransferObject->eState == etwsUSERBUSY) ||
				  (TransferObject->eState == etwsSENDERROR) ) )
			{

				// idle state ... check for timeout to reconnect
				if ( TransferObject->tTimeout == 0 )
				{
					// send connection request to the hub on timeout
					i = pDCLibServerManager->SendConnectionRequest( TransferObject->sNick, TransferObject->sHubName );

					switch (i)
					{
						case 0:
							TransferObject->eState = etwsWAIT;
							break;
						case -1:
							TransferObject->eState = etwsUSEROFFLINE;
							break;
						case -2:
						case -3:
							TransferObject->eState = etwsHUBOFFLINE;
							break;
						case -4:
							TransferObject->eState = etwsSENDERROR;
							break;
						default:
							break;
					}

					TransferObject->tTimeout = ttimeout;
					SendFileInfo( TransferObject );
				}
				else if ( (ttimeout-TransferObject->tTimeout) >= pDCLibConfig->GetTransferResponseTimeout() )
				{
					// set timeout
					TransferObject->tTimeout = 0;
					SendFileInfo( TransferObject );
				}
			}
		}
	}

	UnLock();
}

/** */
void CDownloadManager::Thread( CObject * )
{
	int i;
	time_t ttimeout;

	if ( iShutdownState == 2 )
	{
#ifdef WIN32
		Sleep(100);
#else
		usleep(100);
#endif
		return;
	}

	ttimeout = time(0);

	if ( iShutdownState == 0 )
	{
		// save queue
		i = pDCLibConfig->GetDownloadQueueTime();

		if ( i > 0 )
		{
			i *= 60;

			if ( (ttimeout-tDownloadQueueTimeout) > i )
			{
				SaveQueue();
				tDownloadQueueTimeout = ttimeout;
			}
		}
	}

	// update the transfer list
	UpdateTransferList(ttimeout);

	// update the queue list
	if ( (ttimeout-tUpdateTransferTimeout) >= 1 )
	{
		UpdateQueueList(ttimeout);
	}

	// update timeout
	tUpdateTransferTimeout = ttimeout;

#ifdef WIN32
	Sleep(50);
#else
	usleep(50);
#endif
}

/** callback function */
int CDownloadManager::DM_ListenCallBack( CObject * /*Sender*/, CObject * Object )
{
	int handle = (int)Object;

	if ( handle == -1 )
	{
		return -1;
	}

	CTransfer * Transfer = new CTransfer();

	Transfer->SetTransferID( pDownloadManager->GetNewTransferID() );
	Transfer->SetNick( pDCLibConfig->GetNick() );
	Transfer->SetRate( pDCLibConfig->GetMaxTransferrate() );
	Transfer->SetCallBackFunction( new CCallback<CDownloadManager>( pDownloadManager, &CDownloadManager::DM_TransferCallBack ) );

	if ( Transfer->SetSocket(handle) == 0 )
	{
		pDownloadManager->SendLogInfo("Incoming connection from '"+Transfer->GetHost()+"'\n");

		pDownloadManager->pTransferList->Lock();
		pDownloadManager->pTransferList->Add( Transfer->GetTransferID(), Transfer );
		pDownloadManager->pTransferList->UnLock();
	}
	else
	{
		delete Transfer;
		close(handle);
	}

#ifdef WIN32
	Sleep(250);
#else
	usleep(250);
#endif	
	return 0;
}

/** */
int CDownloadManager::DM_TransferCallBack( CObject * transfer, CObject * Object )
{
	pDownloadManager->TransferCallBackThread.Lock();

	CDCMessage *DCMsg;
	CByteArray ba;
	CString s,r;
	ulonglong  len;
	CDir dir;
	CList<CString> SharedFolders;
	bool remove,bdirec;

	CTransfer * Transfer = (CTransfer*)transfer;

	DCMsg = (CDCMessage*)Object;

	switch ( DCMsg->eType )
	{
		case DC_MESSAGE_MYNICK:
		{
			// check over the nick for download or upload
			if ( pDownloadManager->CheckWaitTransfer( Transfer ) == FALSE )
			{
				Transfer->SetSrcDirection(edUPLOAD);
			}
			else
			{
				Transfer->SetSrcDirection(edDOWNLOAD);
			}

			if ( Transfer->GetSrcDirection() != edNONE )
			{
				Transfer->SendDirection( Transfer->GetSrcDirection(), Transfer->GetLevel() );
			}

			break;
		}

		case DC_MESSAGE_GET:
		{
			// remote want a file ...
			CMessageGet * msg = (CMessageGet*)Object;

			if ( Transfer->GetSrcDirection() == edUPLOAD )
			{
				// send a log info
				pDownloadManager->SendLogInfo( "Upload: " + msg->sFilename, Transfer );

				// check if the filelist
				if ( msg->sFilename == CString(DC_USER_FILELIST) )
				{
					if ( pDCLibConfig->GetShareSize() == 0 )
					{
						Transfer->SendError("File not found.");

						pDownloadManager->SendLogInfo( "Upload (No share): " + msg->sFilename, Transfer );
					}
					else
					{
						if ( pDCLibConfig->GetShareBuffer( &ba ) == 0 )
						{
							// set send buffer
							Transfer->SetBuffer(&ba);
							// set local transfer medium to buffer
							Transfer->SetMedium(eltBUFFER);
							// start upload
							Transfer->StartUpload( msg->sFilename, pDCLibConfig->GetShareBufferSize(), msg->lPos-1, "" );
						}
						else
						{
							Transfer->SendError("File not found.");
						}
					}
				}
				else // remote download from the share ...
				{
					// get the shared folder list
					pDCLibConfig->GetSharedFolders(&SharedFolders);

					// search the wanted file in the share
					if ( (s = dir.GetCorrectFilePath(msg->sFilename.Data(),&SharedFolders)) == "" )
					{
						Transfer->SendError("File not found.");

						pDownloadManager->SendLogInfo( "Upload (File not found): " + msg->sFilename, Transfer );
					}
					else
					{
						// file found, check filesize
						len = dir.getFileSize(s,FALSE);

						if ( msg->lPos >= len ) // if the wanted length >= local length send an error
						{
							Transfer->SendError("File allready download.");
						}
						else if ( msg->lPos == 0 ) // we begin at 1 (see dc protocoll)
						{
							Transfer->SendError("Wrong file position.");
						}
						else // upload the wanted file ...
						{
							// set local transfer medium to buffer
							Transfer->SetMedium(eltFILE);
							// start upload
							Transfer->StartUpload( msg->sFilename, len, msg->lPos-1, s );
						}
					}
				}
			}
			else
			{
				pDownloadManager->SendLogInfo( "Warning wrong mode", Transfer );
				Transfer->Disconnect(TRUE);
			}

			break;
		}

		case DC_MESSAGE_DIRECTION:
		{
			// direction message ...
			CMessageDirection * msg = (CMessageDirection*)Object;
			bdirec = FALSE;

			printf("DIRECTION: level: LOCAL: %d REMOTE: %d\n",Transfer->GetLevel(), msg->iLevel);
			printf("DIRECTION: direc: LOCAL: %d REMOTE: %d\n",Transfer->GetSrcDirection(), msg->Direction);

			// equal direction ...
			if ( msg->Direction == Transfer->GetSrcDirection() )
			{
				// check the level ...
				if ( msg->iLevel < Transfer->GetLevel() )
				{
					// now we must change the dst direction
					if ( Transfer->GetSrcDirection() == edDOWNLOAD )
					{
						Transfer->SetDstDirection(edUPLOAD);
						bdirec = TRUE;
					}
				}
				else if ( msg->iLevel > Transfer->GetLevel() )
				{
					if ( Transfer->GetSrcDirection() == edDOWNLOAD )
					{
						// we change to upload and re-check for free slots
						if ( pDownloadManager->ChangeDirection(Transfer) == TRUE )
						{
							Transfer->SetSrcDirection(edUPLOAD);
							bdirec = TRUE;
						}
					}
				}
			}
			else
			{
				bdirec = TRUE;
			}

			if ( bdirec == TRUE )
			{
				if ( pDownloadManager->SetDirection(Transfer) == FALSE )
				{
					bdirec = FALSE;
				}
			}

			if ( bdirec == FALSE )
			{
				pDownloadManager->SendLogInfo( "Warning no more free slots", Transfer );
				Transfer->SetDstDirection(edNONE);
				Transfer->SendMaxedOut();
				Transfer->Disconnect(TRUE);
			}
			else
			{
				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					printf("DIRECTION: download mode ...\n");

					// next file
					if ( pDownloadManager->SetNextFile(Transfer) == FALSE )
					{
						printf("DIRECTION: download mode without destination file -> disconnecting!\n");
						Transfer->Disconnect(TRUE);
					}
				}
				else if ( Transfer->GetSrcDirection() == edUPLOAD )
				{
					printf("DIRECTION: we are in upload mode ...\n");
				}
			}

			break;
		}

		case DC_MESSAGE_CONNECTION_STATE:
		{
			// connection-state message, gernerated by the ctransfer-class
			remove = TRUE;
			CMessageConnectionState *msg = (CMessageConnectionState*)Object;

			if ( msg->eState == estCONNECTED )
			{
				remove = FALSE;
			}
			else if ( msg->eState == estDISCONNECTED )
			{
				pDownloadManager->SendLogInfo( "Disconnected", Transfer );
			}
			else if ( msg->eState == estSOCKETERROR )
			{
				remove = FALSE;
			}
			else if ( msg->eState == estCONNECTIONTIMEOUT )
			{
				remove = FALSE;
			}

			if ( (remove == TRUE) && (Transfer->GetSrcDirection() == edDOWNLOAD) )
			{
				if ( Transfer->GetMedium() == eltBUFFER )
					Transfer->SetStartPosition(0);
				pDownloadManager->UpdateFileState(Transfer,etfsNONE);
			}

			if ( remove == TRUE )
			{
				Transfer->SetDone(1);
			}

			break;
		}

		case DC_MESSAGE_TRANSFER:
		{
			// transfer message, generated by the ctransfer-class
			CMessageTransfer *msg = (CMessageTransfer*)Object;

			// filetransfer done
			if ( msg->lTransfered == msg->lLength )
			{
				// download ready
				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					pDownloadManager->UpdateFileState(Transfer,etfsNONE);

					// show user file list
					if ( Transfer->GetDstFilename() == CString(DC_USER_FILELIST) )
					{
						pDownloadManager->FileListDone(Transfer);
					}

					if ( msg->lTransfered == Transfer->GetLength() )
					{
						if ( pDownloadManager->SetNextFile(Transfer) == FALSE )
						{
						}
					}
				}
			}
			else 
			{
				// change new chunk end only for file medium
				if ( (Transfer->GetSrcDirection() == edDOWNLOAD) && ( Transfer->GetMedium() == eltFILE ) )
				{
					ulonglong lstart,lend;

					if ( (msg->lLength-msg->lTransfered) < (CHUNK_SIZE/2) )
					{
						if ( pDownloadManager->GetNewChunkEnd( Transfer->GetSrcFilename(), Transfer->GetStartPosition(), Transfer->GetEndPosition(), Transfer->GetCurrentPosition(), &lstart, &lend ) == TRUE )
						{
							Transfer->SetStartPosition(lstart);
							Transfer->SetEndPosition(lend);
						}
					}
				}
			}

			break;
		}

		case DC_MESSAGE_FILELENGTH:
		{
			// remote send a filelength
			CMessageFileLength * msg = (CMessageFileLength*)Object;

			if ( msg->lFileLength == 0 )
			{
				pDownloadManager->SendLogInfo( "Warning: Filelength is NULL.", Transfer );

				if ( Transfer->GetSrcDirection() == edDOWNLOAD )
				{
					pDownloadManager->UpdateFileState(Transfer,etfsERROR);
				}

				Transfer->Disconnect(TRUE);
			}
			else
			{
				pDownloadManager->SendTransferInfo(Transfer);
			}

			break;
		}

		case DC_MESSAGE_MAXEDOUT:
		{
			// no free slots on the remote side
			pDownloadManager->SendLogInfo( "Busy", Transfer );
			Transfer->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_GETLISTLEN:
		{
			// remote want the listlen
			Transfer->SendListLen(pDCLibConfig->GetShareBufferSize());
			break;
		}

		case DC_MESSAGE_ERROR:
		{
			// remote send a error message
			CMessageError * msg = (CMessageError*)Object;

			pDownloadManager->SendLogInfo( "Error: " + msg->sError, Transfer );

			if ( Transfer->GetSrcDirection() == edDOWNLOAD )
			{
				pDownloadManager->UpdateFileState(Transfer,etfsERROR);

				// set next file for download
				if ( pDownloadManager->SetNextFile(Transfer) == FALSE )
				{
				}
			}
			break;
		}

		default:
		{
			printf("dctransfer unknown message: %d\n",DCMsg->eType);
			break;
		}
	}

	delete Object;

	pDownloadManager->TransferCallBackThread.UnLock();

	return 0;
}
