/***************************************************************************
                           cfilemanager.cpp  -  description
                             -------------------
    begin                : Sat Aug 03 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif

#include <dclib/dcos.h>
#include <dclib/cmd5.h>
#include <dclib/cconfig.h>
#include <dclib/cstringlist.h>
#include <dclib/cdir.h>

#include "cfilemanager.h"

CFileManager * pFileManager = 0;

/** */
CFileManager::CFileManager()
{
	pFileList     = new CThreadStringList();
	pTempFileList = 0;

	pDCLibConfig->LoadDCShare( pFileList );
}

/** */
CFileManager::~CFileManager()
{
	Stop();

	Lock();

	if ( pFileList )
	{
		delete pFileList;
		pFileList = 0;
	}

	UnLock();
}

/** */
void CFileManager::SaveIndex()
{
	Lock();
	pDCLibConfig->SaveDCShare( pFileList );
	UnLock();
}

/** */
void CFileManager::StartIndex()
{
	Stop();

	Lock();

	pTempFileList = pFileList;
	pFileList     = new CThreadStringList();

	UnLock();
}

/** */
void CFileManager::StopIndex()
{
	Lock();

	if ( pTempFileList )
	{
		delete pTempFileList;
		pTempFileList = 0;
	}

	UnLock();

	SaveIndex();

//	Start();
}

/** */
void CFileManager::AddIndex( CFileInfo *fileinfo, CString path, CString rootpath )
{
	CString s, *st;
	CFileBaseObject * FileBaseObject, *FileBaseObject1;

	if ( (!pTempFileList) || (!pFileList))
	{
		return;
	}

	Lock();

	s = path + DIRSEPARATOR + fileinfo->name;

	if ( pTempFileList->Get(s,(CObject*&)FileBaseObject) == 0 )
	{
		// file is different (remove old hash)
		if ( (FileBaseObject->lSize != fileinfo->size) ||
		     (FileBaseObject->st_c_time != fileinfo->st_c_time) ||
		     (FileBaseObject->st_m_time != fileinfo->st_m_time) )
		{
			printf("different\n");
			FileBaseObject->sMD4Hash = "";
			FileBaseObject->sMD5Hash = "";
			FileBaseObject->pMD4ChunkHashList.Clear();
		}

		FileBaseObject1 = new CFileBaseObject();

		FileBaseObject1->sMD4Hash = FileBaseObject->sMD4Hash;
		FileBaseObject1->sMD5Hash = FileBaseObject->sMD5Hash;

		st = 0;

		while ( FileBaseObject->pMD4ChunkHashList.Next((CObject*&)st) )
		{
			FileBaseObject->pMD4ChunkHashList.Add(*st,new CString(*st));
		}

		pTempFileList->Del(s);
	}
	else
	{
		printf("not found: %s\n",s.Data());
		FileBaseObject1 = new CFileBaseObject();
	}

	FileBaseObject1->ID        = pFileList->Count()+1;
	FileBaseObject1->lSize     = fileinfo->size;
	FileBaseObject1->sFile     = fileinfo->name;
	FileBaseObject1->sPath     = path;
	FileBaseObject1->sRootPath = rootpath;
	FileBaseObject1->st_a_time  = fileinfo->st_a_time;
	FileBaseObject1->st_c_time  = fileinfo->st_c_time;
	FileBaseObject1->st_m_time  = fileinfo->st_m_time;
	FileBaseObject1->eFileType = eftUNKNOWN;

	pFileList->Add(s,FileBaseObject1);

	UnLock();
}

/** */
void CFileManager::Thread( CObject * )
{
	Lock();

	CFileBaseObject * FileBaseObject = 0;
	CString s;
	FILE * file;
	MD5 * md5;

	while ( pFileList->Next( (CObject*&)FileBaseObject) )
	{
		if ( (FileBaseObject->sMD5Hash == "") &&
		     (FileBaseObject->lSize > 0) )
		{
			// create the hash

			s = FileBaseObject->sRootPath + DIRSEPARATOR + FileBaseObject->sFile;
			printf("CREATE HASH: %s\n",s.Data());

			if ( (file=fopen(s.Data(),"rb")) != 0 )
			{
				md5 = new MD5(file);
				FileBaseObject->sMD5Hash = md5->hex_digest();
				if ( FileBaseObject->sMD5Hash == "" )
					FileBaseObject->sMD5Hash = "ERROR";
				delete md5;
				break;
			}
			else
			{
				printf("ERROR: %s\n",FileBaseObject->sFile.Data());
				FileBaseObject->sMD5Hash = "ERROR";
			}
		}
	}

	UnLock();

	if ( FileBaseObject == 0 )
	{
		Stop();
	}
#ifdef WIN32
	Sleep(250);
#else
	usleep(250);
#endif
}
