/***************************************************************************
                           chttp.cpp  -  description
                             -------------------
    begin                : Mon Jul 29 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include <dclib/dcos.h>

#include "chttp.h"

/** */
CHttp::CHttp()
{
}

/** */
CHttp::~CHttp()
{
}

/** */
int CHttp::GetUrl( CString url )
{
	int i,i1;

	CString s(url);

	Disconnect();

	Stop();

	iErrorCode = 0;
	bData      = FALSE;
	sHeader    = "";
	baData.SetSize(0);

	if ( s == "" )
	{
		return -1;
	}

	CString header = s.Mid(0,7);

	header = header.ToUpper();

	if ( header == "HTTP://" )
	{
		s = s.Mid(7,s.Length()-7);
	}

	if ( (i = s.Find(':')) >= 0 )
	{
		sHost = s.Mid(0,i);
	}

	if ( (i1 = s.Find('/')) < 0 )
	{
		return -1;
	}

	if ( i >= 0 )
	{
		sPort = s.Mid(i+1,i1-i-1);
	}
	else
	{
		sHost = s.Mid(0,i1);
		sPort = "80";
	}

	sUrl = s.Mid(i1,s.Length()-i1);

	Connect( sHost, sPort.asINT() );

	Start();

	return 0;
}

/** */
int CHttp::CallBack_SendObject( CObject * Object )
{
	int err;

	err = DC_CallBack( Object );

	if ( err == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}

	return err;
}

/** */
void CHttp::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->eType = DC_MESSAGE_CONNECTION_STATE;

	Object->eState   = state;
	Object->sMessage = GetSocketError();

	if ( state == estCONNECTED )
	{
		eMode = estTRANSFERHANDSHAKE;
	}
	else if ( state == estDISCONNECTED )
	{
		eMode = estNONE;
		Stop();
	}

	CallBack_SendObject(Object);
}

/** */
void CHttp::DataAvailable( const char * buffer, int len )
{
	int i,i1,i2;

	if ( eMode == estTRANSFERDOWNLOAD )
	{
		if ( bData == FALSE )
		{
			for(i=0;i<(len-4);i++)
			{
				if( (buffer[i]   == 0x0d) &&
				    (buffer[i+1] == 0x0a) &&
				    (buffer[i+2] == 0x0d) &&
				    (buffer[i+3] == 0x0a) )
				{
					sHeader += CString().Set(buffer,i);
					i+=4;

					// parse header ...
					if ( (i1 = sHeader.Find("HTTP/1.1 ")) == -1 )
						Disconnect();
					else if ( (i2 = sHeader.Find(' ',i1+9)) == -1 )
						Disconnect();
					else
					{
						iErrorCode = sHeader.Mid(i1+9,i2-i1-9).asINT();

						if ( iErrorCode == 200 )
						{
							bData = TRUE;
						}
						else
						{
							printf("http error %d\n",iErrorCode);
							Disconnect();
						}
					}

					break;
				}
			}

			if ( (bData == FALSE) && (iErrorCode == 0) )
			{
				sHeader += buffer;
			}
			else if ( iErrorCode != 0 )
			{
				baData.Append((const unsigned char*)(buffer+i),len-i);
			}
		}
		else
		{
			baData.Append((const unsigned char*)buffer,len);
		}
	}
}

/** */
void CHttp::DataSend()
{
	CString header;

	if ( eMode == estTRANSFERHANDSHAKE )
	{
		// send get request
		header  = "GET ";
		header += sUrl;
		header += " HTTP/1.0";
		header += "\xd\xa";
		header += "Host: ";
		header += sHost;
		header += "\xd\xa";
		header += "\xd\xa";

		if ( CSocket::Write( header.Data(), header.Length(), 0, 500 ) != header.Length() )
		{
			eMode = estNONE;

			Disconnect();
		}
		else
		{
			eMode = estTRANSFERDOWNLOAD;
		}
	}
}

/** */
void CHttp::DataTimeout()
{
	Disconnect();
	printf("data timeout\n");
}
