/***************************************************************************
                          chubsearch.cpp  -  description
                             -------------------
    begin                : Fri Mar 15 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cconfig.h>
#include <dclib/cmessagehandler.h>
#include <dclib/cservermanager.h>

#include "chubsearch.h"

extern CConfig * pDCLibConfig;
extern CServerManager * pDCLibServerManager;

CHubSearch * _pHubSearch;

CHubSearch::CHubSearch()
{
	_pHubSearch = this;

	sBuffer = "";
	pHubServerList = 0;

	pHubSearchClientList = new CList<CObject>();

	ManagerThread.SetThreadCallBackFunction( new CCallback<CHubSearch>( this, &CHubSearch::DC_ManagerCallBack ) );
	ManagerThread.Start();
}

CHubSearch::~CHubSearch()
{
	ManagerThread.Stop();
	ManagerThread.SetThreadCallBackFunction(0);

	// clear the client list
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( pHubSearchClientList != 0 )
	{
		while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(0)) != NULL) )
		{
			HubSearchClient->Client->SetCallBackFunction(0);
			HubSearchClient->Client->Disconnect(TRUE);
			delete HubSearchClient->Client;
			HubSearchClient->Client = 0;
			pHubSearchClientList->Remove(HubSearchClient);
			delete HubSearchClient;
		}

		delete pHubSearchClientList;
		pHubSearchClientList = 0;
	}

	HubSearchClientListThread.UnLock();

	// clear hubserverlist
	if ( pHubServerList != 0 )
	{
		delete pHubServerList;
		pHubServerList = 0;
	}
}

/** */
void CHubSearch::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->eType = DC_MESSAGE_CONNECTION_STATE;

	Object->eState   = state;
	Object->sMessage = GetSocketError();

	if ( DC_CallBack( Object ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}
}

/** */
int CHubSearch::HandleMessage( char * c, int )
{
	int p=0;
	eDCMessage type;
	CObject * Object;

	CMessageHandler MessageHandler;

	CString s;

	s = c;

	while( (type=MessageHandler.Parse(s,p,Object=0)) != DC_MESSAGE_PARSE_ERROR )
	{
		switch (type)
		{
			case DC_MESSAGE_SEARCHRESULT:
			{
				break;
			}

			default:
			{
				// remove message
				delete Object;
				Object = 0;
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->eType = type;

			if ( DC_CallBack( Object ) == -1 )
			{
				printf("CallBack failed ...\n");
				delete Object;
			}
		}
	}

	return 0;
}

/** */
void CHubSearch::DataAvailable( const char * buffer, int )
{
	sBuffer = buffer;

	// add '|' for parser
	sBuffer += '|';

	HandleMessage( (char*)sBuffer.Data(), sBuffer.Length() );
}

/** */
int CHubSearch::StartSearch( int mode )
{
	sHubSearchClient * HubSearchClient;

	ehSearchType = ehstNONE;

	// reset current server
	sCurrentServer = 0;

	// set start time
	tStartTime = time(0);

	// clear client list
	HubSearchClientListThread.Lock();

	while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(0)) != NULL) )
	{
		pHubSearchClientList->Remove(HubSearchClient);

		HubSearchClient->Client->SetCallBackFunction(0);
		HubSearchClient->Client->Disconnect();
		delete HubSearchClient->Client;
		HubSearchClient->Client = 0;
		delete HubSearchClient;
	}

	HubSearchClientListThread.UnLock();

	// reset values
	iCurrentHubServer = 0;

	if ( pHubServerList != 0 )
	{
		delete pHubServerList;
		pHubServerList = 0;
	}

	if ( mode == 0 )
	{
		ehSearchType = ehstLOCAL;

		if ( pDCLibServerManager->SendStringToConnectedServers(sSearchString) == 0 )
		{
			ehSearchType = ehstNONE;

			return -1;
		}

		ehSearchType = ehstWAITTIMEOUT;
	}
	else
	{
		if ( mode == 1 )
		{
			// get public hub server list
			pHubServerList = pDCLibServerManager->GetPublicHubServerList();
		}
		else if ( mode == 2 )
		{
			// get bookmark hub server list
			pHubServerList = pDCLibServerManager->GetBookmarkHubServerList();
		}

		if ( (pHubServerList == 0) || (pHubServerList->Count() <= 0) )
		{
			return -1;
		}

		ehSearchType = ehstGLOBAL;
	}

	Start();

	return 0;
}

/** */
void CHubSearch::StopSearch()
{
	ehSearchType = ehstSTOP;

	Disconnect();

	Stop();
}

/** */
void CHubSearch::SendDebug( CString s )
{
	CMessageLog * MessageLog = new CMessageLog();

	MessageLog->sMessage = s;
	MessageLog->eType    = DC_MESSAGE_LOG;

	if ( DC_CallBack( MessageLog ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete MessageLog;
	}
}

/** */
int CHubSearch::DC_ManagerCallBack( CObject *, CObject * )
{
	if ( (_pHubSearch->ehSearchType == ehstGLOBAL) ||
	     (_pHubSearch->ehSearchType == ehstWAITTIMEOUT) ||
	     (_pHubSearch->ehSearchType == ehstSTOP) )
	{
		_pHubSearch->CheckClient();

		if ( _pHubSearch->ehSearchType == ehstGLOBAL )
		{
			_pHubSearch->NewClient();
		}

		if ( _pHubSearch->ehSearchType == ehstSTOP )
		{
			_pHubSearch->HubSearchClientListThread.Lock();

			if ( _pHubSearch->pHubSearchClientList->Count() == 0 )
			{
				_pHubSearch->ehSearchType = ehstNONE;
			}

			_pHubSearch->HubSearchClientListThread.UnLock();
		}
	}

#ifdef WIN32
	Sleep(100);
#else
	usleep(100);
#endif

	return 0;
}

/** */
void CHubSearch::CheckClient()
{
	sHubSearchClient * HubSearchClient,* OldHubSearchClient;

	HubSearchClientListThread.Lock();

	if ( pHubSearchClientList != 0 )
	{
		HubSearchClient = OldHubSearchClient = 0;

		while ( ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next(HubSearchClient)) != NULL) )
		{
			if ( (HubSearchClient->bRemove == TRUE) || ((time(0)-HubSearchClient->timeHubTimeout) >= 300) )
			{
				// check remove timeout
				if ( (HubSearchClient->bRemove == TRUE) && ((time(0)-HubSearchClient->timeHubTimeout) < 6) )
				{
					continue;
				}

				HubSearchClient->Client->SetCallBackFunction(0);

				if ( HubSearchClient->Client->GetConnectionState() != estNONE )
				{
					if ( HubSearchClient->Client->GetConnectionState() != estDISCONNECTING )
					{
						HubSearchClient->Client->Disconnect(TRUE);
					}
					OldHubSearchClient = HubSearchClient;
				}
				else
				{
					delete HubSearchClient->Client;
					HubSearchClient->Client = 0;
					pHubSearchClientList->Remove(HubSearchClient);
					delete HubSearchClient;
					HubSearchClient = OldHubSearchClient;
				}
			}
        	        else
			{
				// remove all clients
				if ( ehSearchType == ehstSTOP )
				{
					HubSearchClient->Client->SetCallBackFunction(0);
					HubSearchClient->bRemove = TRUE;
				}

				OldHubSearchClient = HubSearchClient;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::RemoveClient( CClient * Client )
{
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( pHubSearchClientList != 0 )
	{
		HubSearchClient = 0;

		while ( (HubSearchClient = (sHubSearchClient *)pHubSearchClientList->Next((CObject*)HubSearchClient)) != NULL )
		{
			if ( HubSearchClient->Client == Client )
			{
				HubSearchClient->timeHubTimeout = time(0);
				HubSearchClient->Client->SetCallBackFunction(0);
				HubSearchClient->bRemove = TRUE;
				break;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::NewClient()
{
	int l;
	int i,port;
	CString server,addr,s;
	sHubSearchClient * HubSearchClient;

	if ( pHubServerList == 0 )
	{
		return;
	}

	// done ?
	if ( pHubServerList->Count() == iCurrentHubServer )
	{
		SendDebug("Serverlist done\n");

		ehSearchType = ehstWAITTIMEOUT;

		return;
	}

	HubSearchClientListThread.Lock();

	l = pHubSearchClientList->Count();

	HubSearchClientListThread.UnLock();

	for(;(l<iMaxThreads)&&(iCurrentHubServer<pHubServerList->Count());l++)
	{
		CClient * Client = new CClient();

		HubSearchClient = new sHubSearchClient;

		HubSearchClient->Client = Client;
		HubSearchClient->bRemove = FALSE;
		HubSearchClient->timeHubTimeout = time(0);

		Client->SetNick(pDCLibConfig->GetNick());
		Client->SetComment(pDCLibConfig->GetDescription());
		Client->SetConnectionType(pDCLibConfig->GetSpeed());
		Client->SetEMail(pDCLibConfig->GetEMail());
		Client->SetVersion(VERSION);
		Client->SetShareSize(CString().setNum(pDCLibConfig->GetShareSize()));
		Client->SetMode(pDCLibConfig->GetMode());
		Client->SetHandleUserList(FALSE);
		Client->SetCallBackFunction( new CCallback<CHubSearch>( this, &CHubSearch::DC_ClientCallBack ) );

		if ( pHubServerList->Next((CObject*&)sCurrentServer) != 0 )
		{
			server = *sCurrentServer;

			iCurrentHubServer++;

			SendDebug( "Search on: [" + CString().setNum(iCurrentHubServer) + "] " + server );

			HubSearchClientListThread.Lock();

			pHubSearchClientList->Add((CObject*)HubSearchClient);

			if ( (i=server.Find(':')) != -1 )
			{
				addr = server.Left(i);
				s = server.Mid(i+1,server.Length()-i-1);
				if ( s != "" )
				{
					port = atoi(s.Data());
					Client->Connect(addr,port);
				}
				else
				{
				Client->Connect(addr);
				}
			}
			else
			{
				Client->Connect(server);
			}
		}
		else
		{
			// end on error
			iCurrentHubServer = pHubServerList->Count();
		}

		HubSearchClientListThread.UnLock();
	}
}

/** */
int CHubSearch::DC_ClientCallBack( CObject * Client, CObject * Object )
{
	_pHubSearch->ClientThread.Lock();

	CClient * c = (CClient*)Client;
	CObject * o = Object;

	if ( (c == 0) || (o == 0) )
	{
		_pHubSearch->ClientThread.UnLock();
		return -1;
	}

	CDCMessage *DCMsg;

	DCMsg = (CDCMessage*) o;

	switch ( DCMsg->eType )
	{
		case DC_MESSAGE_CONNECTION_STATE:
		{
			CMessageConnectionState *msg = (CMessageConnectionState*) o;

			switch(msg->eState)
			{
				case estCONNECTED:
				{
					break;
				}

				case estSOCKETERROR:
				{
					_pHubSearch->SendDebug( "SocketError on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
					break;
				}

				case estDISCONNECTED:
				{
					_pHubSearch->RemoveClient(c);
				}

				default:
				{
					break;
				}
			}

			break;
		}

		case DC_MESSAGE_VALIDATEDENIDE:
		{
			_pHubSearch->SendDebug( "Validate denide " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_HUBISFULL:
		{
			_pHubSearch->SendDebug( "Hub is full on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_FORCEMOVE:
		{
			_pHubSearch->SendDebug( "Force move on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_LOCK:
		case DC_MESSAGE_HUBNAME:
		case DC_MESSAGE_CHAT:
		case DC_MESSAGE_MYNICK:
		case DC_MESSAGE_PRIVATECHAT:
		case DC_MESSAGE_SEARCH:
		case DC_MESSAGE_HELLO:
		case DC_MESSAGE_QUIT:
		case DC_MESSAGE_GETINFO:
		{
			break;
		}

		case DC_MESSAGE_MYINFO:
		{
			CString s;
			CMessageMyInfo * MessageMyInfo = (CMessageMyInfo *)o;

			if ( MessageMyInfo->sNick == pDCLibConfig->GetNick() )
			{
				s = _pHubSearch->sSearchString;
				c->SendString(s);
				_pHubSearch->RemoveClient(c);
			}
			break;
		}

		default:
		{
			printf("callback: %d\n",DCMsg->eType);
			break;
		}
	}

	delete o;

	_pHubSearch->ClientThread.UnLock();

	return 0;
}
