/***************************************************************************
                          cquerymanager.cpp  -  description
                             -------------------
    begin                : Sat Jun 8 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include <dclib/dcos.h>
#include <dclib/cdir.h>
#include <dclib/cstringlist.h>
#include <dclib/cservermanager.h>
#include <dclib/cdownloadmanager.h>
#include <dclib/cconfig.h>

#include "cquerymanager.h"

extern CServerManager * pDCLibServerManager;
extern CDownloadManager * pDownloadManager;

/** filetype list */
typedef struct filetype {
	const char* sExt;
	eFileTypes eFileType;
} filetype;

struct filetype FileType[] = {
	// MP3 2
	{ "AIFF", eftMP3 },
	{ "WAV",  eftMP3 },
	{ "AU",   eftMP3 },
	{ "MID",  eftMP3 },
	{ "MP2",  eftMP3 },
	{ "MP3",  eftMP3 },

	// ARCHIVE 3
	{ "ZIP", eftARCHIVE },
	{ "ACE", eftARCHIVE },
	{ "RAR", eftARCHIVE },
	{ "GZ",  eftARCHIVE },
	{ "TAR", eftARCHIVE },
	{ "BZ2", eftARCHIVE },
	{ "RPM", eftARCHIVE },
	{ "CAB", eftARCHIVE },
	{ "EX_", eftARCHIVE },

	// DOCUMENT 4
	{ "PS",    eftDOCUMENT },	
	{ "DOC",   eftDOCUMENT },
	{ "TXT",   eftDOCUMENT },
	{ "PDF",   eftDOCUMENT },
	{ "HTM",   eftDOCUMENT },
	{ "HTML",  eftDOCUMENT },
	{ "SHTML", eftDOCUMENT },

	// APPL 5
	{ "EXE",  eftAPPLICATION },
	{ "DLL",  eftAPPLICATION },
	{ "BAT",  eftAPPLICATION },
	{ "COM",  eftAPPLICATION },
	{ "JS",   eftAPPLICATION },

	// PICTURE 6
	{ "3DS",  eftPICTURE },
	{ "A11",  eftPICTURE },
	{ "ACB",  eftPICTURE },
	{ "ADC",  eftPICTURE },
	{ "ADI",  eftPICTURE },
	{ "AFI",  eftPICTURE },
	{ "AI",   eftPICTURE },
	{ "AIS",  eftPICTURE },
	{ "ANS",  eftPICTURE },
	{ "ART",  eftPICTURE },
	{ "B8",   eftPICTURE },
	{ "CBM",  eftPICTURE },
	{ "PIC",  eftPICTURE },
	{ "PNG",  eftPICTURE },
	{ "TIF",  eftPICTURE },
	{ "PCX",  eftPICTURE },
	{ "BMP",  eftPICTURE },
	{ "GIF",  eftPICTURE },
	{ "JPEG", eftPICTURE },
	{ "JPG",  eftPICTURE },

	// VIDEO 7
	{ "AVI",  eftVIDEO },
	{ "MOV",  eftVIDEO },
	{ "ASF",  eftVIDEO },
	{ "MPG",  eftVIDEO },
	{ "MPEG", eftVIDEO },
	{ "DAT",  eftVIDEO },

	// END
	{ 0, eftALL }
};

const char * BadWord[] = {
	"PORN",
	"PORNO",
	"TEEN",
	"XXX",
	"HARDCORE",
	"HUREN",
	"EDELHUREN",
	"ANAL",
	"SUPERCUM",
	"FUCK",
	"FUCKED",
	"SEXINPUBLIC",
	"LESB",
	"LESBIAN",
	"TITTEN",
	"SPERMA",
	0
};

CQueryManager * pQueryManager = 0;

/** */
CQueryManager::CQueryManager()
{
	pQueryManager = this;

	pQueryQueue = new CList<CQueryObject>;

	pQueryStringIndex = new CStringList();
	pFileBaseList    = new CStringList();

	Start();
}

/** */
CQueryManager::~CQueryManager()
{
	Stop();

	Lock();

	delete pQueryStringIndex;
	delete pFileBaseList;
	delete pQueryQueue;

	UnLock();
}

/** check if the string in the bad word list */
bool CQueryManager::CheckBadWord( CString s )
{
	bool res = FALSE;
	unsigned int i;

	for(i=0;BadWord[i] != 0;i++)
	{
		if ( BadWord[i] == s )
		{
			res = TRUE;
			break;
		}
	}

	return res;
}

/** return the filetype */
eFileTypes CQueryManager::GetFileType( CString ext )
{
	eFileTypes res=eftUNKNOWN;
	unsigned int i;

	for(i=0;FileType[i].sExt != 0;i++)
	{
		if ( FileType[i].sExt == ext )
		{
			res = FileType[i].eFileType;
			break;
		}
	}

	return res;
}

/** */
void CQueryManager::DeleteIndex()
{
	Lock();

	pQueryStringIndex->Clear();
	pFileBaseList->Clear();

	UnLock();
}

/** */
void CQueryManager::AddIndex( CString name, CString path, ulonglong size )
{
	Lock();

	long i;
	CString s,s1;
	CString id;
	CStringList *sl;

	// add object to the base list
	CFileBaseObject * FileBaseObject = new CFileBaseObject();

	FileBaseObject->ID    = pFileBaseList->Count()+1;
	FileBaseObject->sFile = name;
	FileBaseObject->sPath = path;
	FileBaseObject->lSize = size;

	// check filetype
	s = name.ToUpper();

	if ( (i=s.FindRev('.')) != -1 )
	{
		// get extension
		s1 = s.Mid(i+1,s.Length()-i-1);
		FileBaseObject->eFileType = GetFileType(s1);
	}
	else
	{
		FileBaseObject->eFileType = eftALL;
	}

	id = CString().setNum(FileBaseObject->ID);
	pFileBaseList->Add( id, FileBaseObject );

	// add to the string index
	s = path + '/' + name + '/';
	// convert special chars
	s = s.Replace(' ',"/");
	s = s.Replace('.',"/");
	s = s.Replace('-',"/");
	s = s.Replace('_',"/");
	s = s.Replace('(',"/");
	s = s.Replace(')',"/");

	s = CDir().ConvertSeparators(s);
	s = s.ToUpper();

	while( (i=s.Find(DIRSEPARATOR)) != -1 )
	{
		s1 = s.Left(i);
		s  = s.Mid(i+1,s.Length()-i-1);

		i++;

		// max length is 3
		if ( s1.Length() < 3 )
			continue;

		if ( pQueryStringIndex->Get( s1, (CObject*&)sl ) != 0 )
		{
			sl = new CStringList();
			pQueryStringIndex->Add( s1, sl );
		}

		sl->Add( id, new CString(id) );
	}

	UnLock();
}

/** */
void CQueryManager::Thread( CObject * )
{
	CString search,result;
	CString *id;
	CStringList *sl;
	CStringList resultlist;
	CQueryObject * queryobject;
	CFileBaseObject * FileBaseObject;
	CQueryResultObject * QueryResultObject;
	CString s,s1;
	bool badword = FALSE;
	long i;
	int depth;
	int slot_max,slot_free;

	Lock();

	while ( (queryobject = pQueryQueue->Next(0)) != 0 )
	{
		pQueryQueue->Remove(queryobject);

		// query timeout 5 sec.
		if ( (time(0)-queryobject->tTimeout) > 5 )
		{
			delete queryobject;
			queryobject = 0;
		}
		else
		{
			break;
		}
	}

	if (queryobject)
	{
		slot_max  = pDCLibConfig->GetMaxUpload();
		slot_free = pDownloadManager->GetUsedSlots();

		// unlimit slots
		if ( slot_max == 0 )
		{
			slot_max = 99+slot_free;
		}

		// check for extra slots
		if ( slot_free > slot_max )
			slot_free = 0;
		else
			slot_free = slot_max-slot_free;

		search = queryobject->pMessageSearch->sString.ToUpper();

		// convert special chars
		search = search.Replace(' '," ");
		search = search.Replace('.'," ");
		search = search.Replace('-'," ");
		search = search.Replace('_'," ");
		search = search.Replace('('," ");
		search = search.Replace(')'," ");

//		printf("querymanager search for '%s' %ld\n",search.Data(),pQueryStringIndex->Count());

		s = search + " ";

		depth = 0;

		while( (i=s.Find(' ')) != -1 )
		{
			s1 = s.Left(i);
			s  = s.Mid(i+1,s.Length()-i-1);

			i++;

			// max length is 3
			if ( s1.Length() < 3 )
				continue;

			// don't search extensions
			if ( GetFileType(s1) != 1 )
				continue;

			// check for bad words and stop the search if true
			if ( CheckBadWord(s1) == TRUE )
			{
//				printf("Found bad word: %s\n",s1.Data());
				resultlist.Clear();
				badword = TRUE;
				break;
			}

			depth++;

			if ( pQueryStringIndex->Get( s1, (CObject*&)sl ) == 0 )
			{
				id = 0;

				while( sl->Next( (CObject*&)id ) != 0 )
				{
					if ( resultlist.Get( *id, (CObject*&)QueryResultObject ) != 0 )
					{
						QueryResultObject = new CQueryResultObject();
						QueryResultObject->ID = *id;
						QueryResultObject->iDepth = 1;
						resultlist.Add( *id, QueryResultObject );
					}
					else
					{
						QueryResultObject->iDepth++;
					}
				}
			}
		}

		if ( badword == TRUE )
		{
			// *ggg*
			if ( queryobject->pMessageSearch->bLocal == TRUE )
			{
/*
				result  = "$SR ";
				result += queryobject->sNick + " ";
				result += "http://www.church.com/\x5";
				result += "666 ";
				result += "666/666";
				result += "\x5";
				result += queryobject->sHubName + " ";
				result += "(" + queryobject->sHubHost + ")\x5";
				result += queryobject->pMessageSearch->sSource + "|";

				pDCLibServerManager->SendStringToConnectedServers(result,queryobject->sHubName);
*/
			}
		}
		else if ( resultlist.Count() > 0 )
		{
			// result found ...

			id = 0;

			while( resultlist.Next( (CObject*&)QueryResultObject ) != 0 )
			{
				// check for a good result
				if ( QueryResultObject->iDepth < ((depth/2)+1) )
				{
					continue;
				}

				if ( pFileBaseList->Get( QueryResultObject->ID, (CObject*&)FileBaseObject ) == 0 )
				{
					// check size
					if ( queryobject->pMessageSearch->bSizeLimit == TRUE )
					{
						if ( queryobject->pMessageSearch->bSizeAtMost == TRUE )
						{
							if ( FileBaseObject->lSize > queryobject->pMessageSearch->lSize )
							{
								continue;
							}
						}
						else
						{
							if ( FileBaseObject->lSize < queryobject->pMessageSearch->lSize )
							{
								continue;
							}
						}
					}

					// check filetype
					switch(queryobject->pMessageSearch->eFileType)
					{
						case eftALL:
							break;
						case eftMP3:
						case eftARCHIVE:
						case eftDOCUMENT:
						case eftAPPLICATION:
						case eftPICTURE:
						case eftVIDEO:
							if ( queryobject->pMessageSearch->eFileType != FileBaseObject->eFileType )
							{
								continue;
							}
							break;
						case eftFOLDER:
							continue;
							break;
						default:
							continue;
							break;
					}

					// result found ! send it ...
					if ( queryobject->pMessageSearch->bLocal == TRUE )
					{
						CString s;
						s = FileBaseObject->sPath + '/' + FileBaseObject->sFile;
						s = s.Replace('/',"\\");
						result  = "$SR ";
						result += queryobject->sNick + " ";
						result += s + "\x5";
						result += CString().setNum(FileBaseObject->lSize) + " ";
						result += CString().setNum(slot_free) + "/" + CString().setNum(slot_max);
						result += "\x5";
						result += queryobject->sHubName + " ";
						result += "(" + queryobject->sHubHost + ")\x5";
						result += queryobject->pMessageSearch->sSource + "|";

						if ( pDCLibServerManager->SendStringToConnectedServers(result,queryobject->sHubName) == 0 )
							break;

//						printf("'%s' '%s' %Lu\n", FileBaseObject->sPath.Data(), FileBaseObject->sFile.Data(), FileBaseObject->lSize );
//						printf("'%s'\n",result.Data());
					}
				}
			}
		}

		delete queryobject;
	}

	UnLock();

#ifdef WIN32
	Sleep(250);
#else
	usleep(250);
#endif
}

bool CQueryManager::SearchQuery( CString hubname, CString hubhost, CString nick, CMessageSearch * msg )
{
	Lock();

	bool res = FALSE;

	CQueryObject * queryobject;

	if ( pQueryQueue->Count() < 20 )
	{
		if ( (msg != 0) && (hubname != "") && (hubhost != "") )
		{
			queryobject = new CQueryObject();

			queryobject->sHubName = hubname;
			queryobject->sHubHost = hubhost;
			queryobject->sNick    = nick;
			queryobject->tTimeout = time(0);
			queryobject->pMessageSearch = new CMessageSearch(msg);
			pQueryQueue->Add(queryobject);

			res = TRUE;
		}
	}

	UnLock();

	return res;
}
